__package__    = "mtdb"
__version__    = "0.1"
__author__     = "Aaron Straup Cope"
__url__        = "http://www.aaronland.info/python/mtdb/"
__cvsversion__ = "$Revision: 1.19 $"
__date__       = "$Date: 2007/09/01 06:28:48 $"
__copyright__  = "Copyright (c) 2007 Aaron Straup Cope. Perl Artistic License."

"""machine tag database

import mtdb
import sqlite3

dbconn = sqlite3.connect('/path/to/mt.db')
store = mtdb.mtdb(dbconn, 0)

uri = "http://www.flickr.com/photos/bar"
store.add_document(uri, ["flickr:user=bar", "dc:age=20"])

for ns in store.find_namespaces() :
    print ns

    for pred in store.find_predicates(ns) :
        print "\t%s:%s" % (ns, pred)

        for v in store.find_values(ns, pred) :
            print "\t\t%s:%s=%s" % (ns, pred, v)


for doc in store.find_documents_by_range((19, 40)) :
    print doc

    for t in store.find_tags(doc) :
        print "\t%s" % t

for doc in store.find_documents_by_like("str%") :
    print doc

    for t in store.find_tags(doc) :
        print "\t%s" % t


for t in store.find_tags(uri) :
    print t

docs = store.find_documents("flickr")

if docs :
    for d in docs :
        print d 

"""

import tagstore
import documentstore
import machinetagstore

import machinetag
import sqlite3

class mtdb :

    """Class used to instantiate a machine tag database"""
    
    def __init__ (self, dbconn, ctx=0) :
        """dbconn is an active database connection conforming to PEP 249 (The
        Python Database API Specification v2.0)

        (I say that but as of version 0.1 this package has only been tested with
        sqlite3. Specifically hooks to trap "integrity errors" for other RDBMS
        systems are missing. Patches are not only welcomed but encouraged.)
        
        ctx is an optional integer representing the 'context' that a machine tag
        is stored with. Context is left up to individual applications to define.
        For example, a context could be a userid or a privacy indicator (friends,
        family, etc.)

        If defined, it is applied to all queries. By default it is left empty.
        """

        self.__dbconn = dbconn;
        self.__context = ctx
        
        self.__tagstore = tagstore.tagstore(dbconn)
        self.__documentstore = documentstore.documentstore(dbconn)
        self.__machinetagstore = machinetagstore.machinetagstore(dbconn)

    #

    def find_namespaces (self, pred=None, value=None) :
        """Return a list containing distinct namespaces in the mtdb.

        pred is an optional string representing a predicate used to filter
        the query. For example, all the namespaces where the predicate is
        'title'.

        value is an optional string representing a value used to filter
        the query. For example, all the namespaces where the predicate is
        'title' and the value is 'OH HAI'.
        """
            
        triples = self.__triples(None, pred, value, False)

        if not triples :
            return []
        
        cur = self.__machinetagstore.namespaces(triples[1], triples[2], self.__context)
        return self.__expand_tags(cur)

    #
    
    def find_predicates (self, ns=None, value=None) :
        """Return a list containing all the distinct predicates in the mtdb.

        ns is an optional string representing a namespace used to filter
        the query. For example, all the predicates where the namespace is
        'dc'.

        value is an optional string representing a value used to filter
        the query. For example, all the predicates where the namespace is
        'flickr' and the value is 'bike'.
        """
        
        triples = self.__triples(ns, None, value, False)

        if not triples :
            return []

        cur = self.__machinetagstore.predicates(triples[0], triples[2], self.__context)
        return self.__expand_tags(cur)

    #
    
    def find_values (self, ns=None, pred=None) :
        """Return a list containing all the distinct values in the mtdb.

        ns is an optional string representing a namespace used to filter
        the query. For example, all the values where the namespace is
        'dc'.

        pred is an optional string representing a predicate used to filter
        the query. For example, all the values where the namespace is
        'flickr' and the predicate is 'user'.
        """
        
        triples = self.__triples(ns, pred, None, False)        

        if not triples :
            return []

        cur = self.__machinetagstore.values(triples[0], triples[1], self.__context)
        return self.__expand_tags(cur)

    #
    
    def __expand_tags (self, cursor=None) :
        """La la la. I can't hear you"""
        
        tags = []

        if not cursor :
            return tags
        
        for row in cursor :
            ns = self.__tagstore.retrieve_by_id(row[0])
            tags.append(ns[1])

        return tags

    #
    
    def find_tags (self, uri) :
        """Return a list of machine tags for a given URI."""
        
        expand = True
        
        doc = self.__documentstore.retrieve(uri, expand)

        if not doc :
            return []

        cur = self.__machinetagstore.retrieve_by_document(doc.id, self.__context)
        tags = []

        for row in cur :

            try :
                ns = self.__tagstore.retrieve_by_id(row[0], expand)
                pred = self.__tagstore.retrieve_by_id(row[1], expand)
                value = self.__tagstore.retrieve_by_id(row[2], expand)
            except Exception, e :
                continue
                
            mt = machinetag.machinetag(ns.as_string(), pred.as_string(), value.as_string())

            if mt.is_machinetag() :
                tags.append(mt)

        return tags
    
    #
    
    def find_documents (self, ns=None, pred=None, value=None) :
        """Returns a list of document URIs matching at least one part of a machine
        tag triple.

        ns is an optional string representing a namespace used to filter
        the query. For example, all the documents where the namespace is
        'dc'.

        pred is an optional string representing a predicate used to filter
        the query. For example, all the documents where the namespace is
        'dc' and the predicate is 'subject'.

        value is an optional string representing a value used to filter
        the query. For example, all the documents where the namespace is
        'dc' and the predicate is 'subject' and the the value is 'ponies'.
        """
        
        triples = self.__triples(ns, pred, value)

        if not triples :
            return []
        
        cur = self.__machinetagstore.retrieve_documents_by_machinetag(triples[0], triples[1], triples[2], self.__context)
        return self.__expand_documents(cur)
    
    #
    
    def find_documents_by_range (self, range, ns=None, pred=None) :
        """Returns a list of document URIs whose numeric value is between a specific
        range.

        value is a list containing two numeric values used to filter the
        query. For example, all the documents where the value is between
        18 and 14.
        
        ns is an optional string representing a namespace used to filter
        the query. For example, all the documents where the value is between
        18 and 14 and the namespace is 'person'.

        pred is an optional string representing a predicate used to filter
        the query. For example, all the documents where the value is between
        18 and 14 and the namespace is 'person' and the predicate is 'age'..
        """
        
        triples = self.__triples(ns, pred, None, False)

        if not triples :
            return []
        
        cur = self.__machinetagstore.retrieve_documents_by_machinetag_range(range, triples[0], triples[1], self.__context) 
        return self.__expand_documents(cur)

    #
    
    def find_documents_by_like (self, like, ns=None, pred=None) :
        """Returns a list of document URIs whose string value matches a given
        string.

        like is a string used to filter the query. For example, all the documents
        where the value starts with 'str'
        
        ns is an optional string representing a namespace used to filter
        the query. For example, all the documents where the value starts with
        'str' and the namespace is 'dc'.
        
        pred is an optional string representing a predicate used to filter
        the query.  For example, all the documents where the value starts with
        'str' and the predicate is 'user'.
        """
        
        triples = self.__triples(ns, pred, None, False)

        if not triples :
            return []

        cur = self.__machinetagstore.retrieve_documents_by_machinetag_like(like, triples[0], triples[1], self.__context) 
        return self.__expand_documents(cur)

    #
    
    def __expand_documents (self, cursor=None) :
        """La la la. I can't hear you"""
        
        docs = []

        if not cursor :
            return []
        
        expand = True
        
        for row in cursor :            
            doc = self.__documentstore.retrieve_by_id(row[3], expand)
            docs.append(doc.as_string())

        return docs
        
    #
    
    def add_document (self, uri, tags) :
        """Associate a document URI with a list of machine tags"""        
        
        for t in tags :
            self.add_tag(uri, t)

    #
    
    def add_tag (self, uri, tag) :
        """Associate a document URI with a single machine tag"""
        
        mt = machinetag.machinetag(tag)
        
        if not mt.is_machinetag() :
            return None

        doc = self.__store_document(uri)

        ns = self.__store_tag(mt.namespace())
        pred = self.__store_tag(mt.predicate())
        value = self.__store_tag(mt.value()) 
                
        return self.__store_machinetag(mt, ns.id, pred.id, value.id, doc.id)

    #

    def remove_tag (self, uri, ns=None, pred=None, value=None) :
        """Remove a machinetag from a document URI

        ns is an optional string representing a namespace with which to
        remove tags. For example, remove all tags whose namespace is 'dc'.

        pred is an optional string representing a predicate with which to
        remove tags. For example, remove all tags whose namespace is 'dc'
        and whose predicate is 'date'.

        value is an optional string representing a predicate with which to
        remove tags. For example, remove all tags whose value is 'poop'.
        
        Note : As of version 0.1 neither the document URI or the individual
        parts of a machine tag are removed from the 'documentstore' or the
        'tagstore', respectively. Only the association between a given URI
        and a machine tag (and its application context) are removed.
        """
        
        row = self.__documentstore.retrieve(uri)

        if not row :
            return False

        triples = self.__triples(ns, pred, value)

        if not triples :
            return False
        
        return self.__machinetagstore.remove_tag(row[0], triples[0], triples[1], triples[2], self.__context)
                              
    #

    def remove_document (self, uri) :
        """Remove all instances of a  document URI from the machine tag store

        Note : As of version 0.1 neither the document URI or the individual
        parts of a machine tag are removed from the 'documentstore' or the
        'tagstore', respectively. Only the association between a given URI
        and a machine tag (and its application context) are removed.
        """

        row = self.__documentstore.retrieve(uri)

        if not row :
            return False
        
        return self.__machinetagstore.remove_document(row[0], self.__context)
        
    #
    
    def __store_tag (self, tag_raw) :
        """La la la. I can't hear you"""
                
        try : 
            self.__tagstore.store(tag_raw)
        except sqlite3.IntegrityError, e :
            pass

        # Please for you to tell me your rdbms' equivalent....
        
        except Exception, e :
            raise Exception(e)

        row = self.__tagstore.retrieve(tag_raw)

        if row : 
            return tagstore.tag(row)

    #
    
    def __store_document (self, doc_uri) :
        """La la la. I can't hear you"""
        
        try : 
            self.__documentstore.store(doc_uri)
        except sqlite3.IntegrityError, e :
            pass

        # Please for you to tell me your rdbms' equivalent....
        
        except Exception, e :
            raise Exception(e)

        row = self.__documentstore.retrieve(doc_uri)

        if row :
            return documentstore.document(row)

    #

    def __store_machinetag (self, mt, ns_id, pred_id, value_id, doc_id) :
        """La la la. I can't hear you"""
        
        try :
            self.__machinetagstore.store(ns_id, pred_id, value_id, doc_id, self.__context, mt.value())
        except sqlite3.IntegrityError, e :
            pass

        # Please for you to tell me your rdbms' equivalent....
        
        except Exception, e :
            raise Exception(e)
        
        return self.__machinetagstore.retrieve_by_machinetag(ns_id, pred_id, value_id, self.__context)

    #
    
    def __triples (self, ns=None, pred=None, value=None, test=True) :
        """La la la. I can't hear you"""
        
        foo = {'ns_id':None, 'pred_id':None, 'value_id':None}
        
        if ns :
            row = self.__tagstore.retrieve(ns)
            if row :
                foo['ns_id'] = row[0]
            else :
                return None
            
        if pred :
            row = self.__tagstore.retrieve(pred)
            if row :
                foo['pred_id'] = row[0]
            else :
                return None
            
        if value :
            row = self.__tagstore.retrieve(value)
            if row :
                foo['value_id'] = row[0]
            else :
                return None
            
        res = [foo['ns_id'], foo['pred_id'], foo['value_id']]
        
        if not test :
            return res
        
        for i in res :
            if i :
                return res

        return None
